import asyncio
import multiprocess as multiprocessing
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from cascata import Graph, inport, outport, component, persist


@component
@inport("count", default=5)
@outport("o")
async def Prod(count, o):
    for i in range(count):
        await o.send(i)


@component
@inport("items")
@persist("store", lambda lst: lst)
async def Collect(items, store):
    store.get().append(items)


@component
@inport("a")
@inport("b")
@persist("store", lambda lst: lst)
async def Pair(a, b, store):
    store.get().append((a, b))


def test_batch_all():
    with multiprocessing.Manager() as manager:
        result = manager.list()
        g = Graph()
        g.p = Prod
        g.c = Collect
        g.c.store < result
        g.p.o >>= g.c.items
        g.run(1)
        assert list(result) == [list(range(5))]


def test_batch_size_two():
    with multiprocessing.Manager() as manager:
        result = manager.list()
        g = Graph()
        g.p = Prod
        g.c = Collect
        g.c.store < result
        g.p.o >> 2 >= g.c.items
        g.run(1)
        assert list(result) == [[0, 1], [2, 3], [4]]


def test_group_with_batch():
    with multiprocessing.Manager() as manager:
        result = manager.list()
        g = Graph()
        g.p1 = Prod
        g.p2 = Prod
        g.p2.count < 4
        g.pair = Pair
        g.pair.store < result
        g.p1.o >> g.pair.a
        g.p2.o >> 2 >= g.pair.b
        g.pair.sync(g.pair.a, g.pair.b)
        g.run(1)
        assert list(result) == [(0, [0, 1]), (1, [2, 3])]

def test_group_sends_batch():
    with multiprocessing.Manager() as manager:
        result = manager.list()
        g = Graph()
        g.prods = Prod * 2
        g.prods.initialize('count', 1)
        g.collect = Collect
        g.collect.store < result
        g.prods.o >> 2 >= g.collect.items
        g.run(1)
        assert list(result) == [[0, 0]]



import asyncio
import multiprocessing
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from cascata.channel import Channel

# Helper functions for multiprocessing

def producer_proc(channel, start, end):
    async def _run():
        async with channel.open():
            for i in range(start, end):
                await channel.put(i)
    asyncio.run(_run())

def consumer_proc(channel, queue):
    async def _run():
        items = []
        async for item in channel:
            items.append(item)
        queue.put(items)
    asyncio.run(_run())

def test_channel_multi_process():
    channel = Channel(capacity=10000)
    q1 = multiprocessing.Queue()
    q2 = multiprocessing.Queue()

    p_prod1 = multiprocessing.Process(target=producer_proc, args=(channel, 0, 500))
    p_prod2 = multiprocessing.Process(target=producer_proc, args=(channel, 500, 1000))
    p_cons1 = multiprocessing.Process(target=consumer_proc, args=(channel, q1))
    p_cons2 = multiprocessing.Process(target=consumer_proc, args=(channel, q2))

    # Start consumers first so they wait on the channel
    p_cons1.start()
    p_cons2.start()
    p_prod1.start()
    p_prod2.start()

    p_prod1.join()
    p_prod2.join()
    p_cons1.join()
    p_cons2.join()

    items1 = q1.get()
    items2 = q2.get()

    assert len(items1) == 1000
    assert len(items2) == 1000
    assert sorted(items1) == list(range(1000))
    assert sorted(items2) == list(range(1000))



import asyncio
from .channel import Channel
from contextlib import asynccontextmanager, AsyncExitStack


class InputPort:
    def __init__(self, name, capacity=100, default=None):
        """
        Initializes an InputPort.

        :param name: The name of the InputPort.
        :param initialization_value: The initial value for the port, if any.
        """
        self.name = name
        self.component = None
        self.capacity = capacity
        self.channel = Channel(capacity)  # Initialize a Channel object
        self.initialization_value = default
        self.batch_size = None
        self.type = None

    def __repr__(self):
        return f"inport {self.component.name}.{self.name}"

    async def put(self, item):
        """
        Puts an item into the channel.

        :param item: The item to be put into the channel.
        """
        await self.channel.put(item)

    @asynccontextmanager
    async def open(self):
        """
        Context manager for managing the state of the channel.
        """
        async with self.channel.open() as channel:
            yield channel

    def initialize(self, value):
        """
        initialize this port with a value
        """
        self.initialization_value = value

    def __aiter__(self):
        """
        Makes InputPort an asynchronous iterable.
        If batch_size is None, items are yielded individually.
        Otherwise, items are collected and yielded as lists.
        """
        if self.batch_size is None:
            return self.channel.__aiter__()

        async def _gen():
            batch = []
            async for item in self.channel:
                batch.append(item)
                if self.batch_size and len(batch) >= self.batch_size:
                    yield batch
                    batch = []
            if batch:
                yield batch

        return _gen()

    def __lt__(self, b):
        self.initialize(b)


class OutputPort:
    def __init__(self, name):
        """
        Initializes an OutputPort.

        :param name: The name of the OutputPort.
        """
        self.name = name
        self.component = None
        self.connections = set()  # List to store connections to InputPorts
        self.type = None

    @asynccontextmanager
    async def open(self):
        async with AsyncExitStack() as stack:
            await asyncio.gather(
                *[stack.enter_async_context(port.open()) for port in self.connections]
            )
            yield

    def __repr__(self):
        return f"inport {self.component.name}.{self.name}"

    async def send(self, item):
        """
        Concurrently sends an item to all connected InputPorts.

        :param item: The item to be sent.
        """
        await asyncio.gather(*(connection.put(item) for connection in self.connections))

    def connect(self, inport):
        """
        Connects this OutputPort to an InputPort.

        :param inport: The InputPort to connect to.
        """
        inport.initialization_value = None
        if getattr(self, "type", None) is not None and getattr(inport, "type", None) is not None:
            if self.type != inport.type:
                raise TypeError(
                    f"Cannot connect {self} ({self.type}) to {inport} ({inport.type})"
                )
        self.component.graph.edge(self, inport)
        if type(inport) is InputPort:
            self.connections.add(inport)

    def connect_batch(self, inport, size=0):
        """Connect with batching semantics."""
        inport.batch_size = size
        self.connect(inport)

    def __rshift__(self, other):
        if isinstance(other, int):
            return _BatchBuilder(self, other)
        self.connect(other)
        return self

    def __irshift__(self, inport):
        self.connect_batch(inport, 0)
        return self


class _BatchBuilder:
    def __init__(self, outport, size):
        self.outport = outport
        self.size = size

    # allow syntax outport >> n >= inport
    def __ge__(self, inport):
        self.outport.connect_batch(inport, self.size)
        return self.outport


class _GroupBatchBuilder:
    """Helper for batching connections from ComponentGroup ports."""

    def __init__(self, handler, size):
        self.handler = handler
        self.size = size

    def __ge__(self, inport):
        self.handler.component.connect_batch(self.handler, inport, self.size)
        return self.handler


class PortHandler:
    def __init__(self, name, parent):
        self.name = name
        self.component = parent

    def __lt__(self, b):
        self.component.initialize(self.name, b)

    def __rshift__(self, other):
        if isinstance(other, int):
            return _GroupBatchBuilder(self, other)
        self.component.connect(self, other)
        return self

    def __irshift__(self, inport):
        self.component.connect_batch(self, inport, 0)
        return self


class PersistentValue:
    def __init__(self, initializer, *args, **kwargs):
        """Holds mutable state initialized once per process."""
        self.initializer = initializer
        self._args = args
        self._kwargs = kwargs
        self.value = None
        self._initialized = False

    def set_args(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._initialized = False

    def __lt__(self, other):
        if isinstance(other, tuple):
            if len(other) == 2 and isinstance(other[1], dict):
                args, kwargs = other
                self.set_args(*args, **kwargs)
            else:
                self.set_args(*other)
        elif isinstance(other, dict):
            self.set_args(**other)
        else:
            self.set_args(other)

    def get(self):
        if not self._initialized:
            self.value = self.initializer(*self._args, **self._kwargs)
            self._initialized = True
        return self.value

    def set(self, value):
        self.value = value
        self._initialized = True



from cascata import inport, outport, component


@component
@inport('x')
@outport('y')
async def echo(x, y):
    await y.send(x)


def test_component_test_basic():
    result = echo.test(x=5)
    assert result == {'y': [5]}


@component
@inport('a')
@outport('b')
@outport('c')
async def multi(a, b, c):
    await b.send(a)
    await c.send(a * 2)


def test_component_test_multiple_outputs():
    result = multi.test(a=3)
    assert result == {'b': [3], 'c': [6]}



import pytest
from cascata import Graph, component, inport, outport

@component
@outport('out')
async def IntSource(out: int):
    await out.send(1)

@component
@inport('inp')
async def IntSink(inp: int):
    pass

@component
@inport('inp')
async def StrSink(inp: str):
    pass

@component
@inport('inp')
async def AnySink(inp):
    pass

def test_type_mismatch_connection():
    g = Graph()
    g.src = IntSource
    g.sink = StrSink
    with pytest.raises(TypeError):
        g.src.out >> g.sink.inp

def test_type_match_connection():
    g = Graph()
    g.src = IntSource
    g.sink = IntSink
    g.src.out >> g.sink.inp  # no error expected

def test_missing_annotation_connection():
    g = Graph()
    g.src = IntSource
    g.sink = AnySink
    g.src.out >> g.sink.inp  # allowed without annotations



import os
import sys
import json

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from cascata import Graph, component, inport, outport



@component
@outport('o')
async def Prod(o):
    await o.send(1)
Prod.__module__ = __name__

@component
@inport('i', capacity=5, default=7)
async def Cons(i):
    pass
Cons.__module__ = __name__


def build_graph():
    g = Graph()
    g.p = Prod
    g.consumers = Cons * 2
    g.consumers.initialize('i', 7)
    g.p.o >> g.consumers.i
    for c in g.consumers.group:
        c.i.batch_size = 2
    g.extra = Cons
    g.extra.i.initialize(42)
    return g


def test_graph_to_from_json_roundtrip():
    g1 = build_graph()
    data = g1.to_json()
    # ensure json serializable
    json.dumps(data)

    # grouped components should serialize once with count
    assert 'consumers_0' not in data['nodes']
    assert 'consumers' in data['group_handles']
    ginfo = data['group_handles']['consumers']
    assert ginfo['count'] == 2
    assert ginfo['inports']['i']['capacity'] == 5
    assert ginfo['inports']['i']['batch_size'] == 2

    # group connectors should not be serialized
    assert all(info['class'] != 'GroupConnector' for info in data['nodes'].values())
    assert data['edges'] == [['p', 'o', 'consumers', 'i']]


    g2 = Graph.from_json(data)

    assert set(g1.nodes.keys()) == set(g2.nodes.keys())
    assert set(g1.group_handles.keys()) == set(g2.group_handles.keys())

    # check port properties preserved
    c0 = g2.consumers.group[0]
    assert c0.i.channel.capacity == 5
    assert c0.i.batch_size == 2
    assert g2.extra.i.initialization_value == 42

    edges1 = sorted([(o.component.name,o.name,i.component.name,i.name) for o,i in g1.edges])
    edges2 = sorted([(o.component.name,o.name,i.component.name,i.name) for o,i in g2.edges])
    assert edges1 == edges2

    # round trip serialization should be stable
    assert g1.to_json() == Graph.from_json(g1.to_json()).to_json()

    g1.run()
    g2.run()


def test_subgraph_group_serialization():
    sg = Graph()
    sg.prods = Prod * 3
    sg.consumers = Cons * 3
    for c in sg.consumers.group:
        c.i.batch_size = 4
    sg.prods.o >> sg.consumers.i

    data = sg.to_json()
    assert 'consumers_0' not in data['nodes']
    assert data['group_handles']['consumers']['count'] == 3
    assert data['group_handles']['consumers']['inports']['i']['batch_size'] == 4
    assert all(info['class'] != 'GroupConnector' for info in data['nodes'].values())
    assert data['edges'] == [['prods', 'o', 'consumers', 'i']]

    sg2 = Graph.from_json(data)
    assert set(sg.nodes.keys()) == set(sg2.nodes.keys())
    c0 = sg2.consumers.group[0]
    assert c0.i.batch_size == 4

    sg.run()
    sg2.run()



import copy
import networkx as nx
import multiprocess
import asyncio
import inspect
from .component import Component, ComponentGroup, GroupConnector
from .port import InputPort, OutputPort
from .log import log
import traceback
try:
    from graphviz import Digraph
except:
    Digraph = False

class DeadlockError(Exception):
    """
    Exception raised when deadlock conditions are detected in a Cascata Graph.

    Attributes:
        issues (list[str]): A list of descriptive deadlock issues found.
    """
    def __init__(self, issues: list[str]):
        self.issues = issues
        message = "Deadlock detected:\n" + "\n".join(f"- {issue}" for issue in issues)
        super().__init__(message)


class Graph:
    """Summary

    Attributes:
        nodes (dict): mapping of names to Component instances
        edges (list): list of tuples (OutputPort, InputPort)
        processes (list): list of processes spawned by run() or start()
        networkx_graph (networkx.DiGraph): internal graph for scheduling
        _local_namespace (dict): namespace where Graph was created
        _exports (dict): mapping of export name to Port
    """
    def __init__(self):
        self.nodes = {}
        self.edges = []
        self.processes = []
        self.networkx_graph = nx.DiGraph()
        self._local_namespace = inspect.currentframe().f_back.f_locals
        self._exports = {}
        self.is_subgraph = False
        self.parent = None
        self.group_handles: dict[str,ComponentGroup] = {}

    def __setattr__(self, name, value):
        # detect adding a subgraph instance
        if isinstance(value, Graph):
            self._add_subgraph(name, value)
        # detect adding a Component subclass or instance
        elif isinstance(value, type) and issubclass(value, Component):
            comp = value(name)
            self.node(name, comp)
        elif isinstance(value, Component):
            self.node(name, value)
        elif isinstance(value, ComponentGroup):
            return self._add_component_group(name, value)            
        else:
            super().__setattr__(name, value)

    def __getattr__(self,name):
        if name in self._exports:
            return self._exports[name]
        if name in self.group_handles:
            return self.group_handles[name]            
        else:
            raise AttributeError(f"Graph object has no attribute {name}")

    def node(self, name, component):
        """Adds a named component to the graph."""
        component.graph = self
        component.name = name
        self.nodes[name] = component
        self.__dict__[name] = component
        self._local_namespace[name] = component

    def edge(self, outport, inport):
        """Connects an outport to an inport."""

        if type(inport.component) is ComponentGroup:
            inport.component.connect(outport, inport)
            return
        root = self._find_root_graph((inport,outport))
        root.edges.append((outport, inport))
        root.networkx_graph.add_edge(outport.component, inport.component)
    
    def _add_component_group(self, alias: str, factory: ComponentGroup):
        """
        Instantiate factory.count copies of factory.comp_cls,
        register them under names alias_0…alias_{n-1},
        record them in self.group_handles and networkx_graph.
        """
        factory.name  = alias
        factory.graph = self

        # 1) instantiate all members
        for i in range(factory.count):
            comp_name = f"{alias}_{i}"
            comp = factory.comp_cls(comp_name)
            comp.name = comp_name
            comp.group_name = alias
            comp.graph = self
            self.nodes[comp_name] = comp
            self.__dict__[comp_name] = comp

            self.networkx_graph.add_node(comp, group=(alias, i, factory.count))

            factory.group.append(comp)
        # 2) remember the group handle
        self.group_handles[alias] = factory

        # 3) expose the alias in local namespace
        self._local_namespace[alias] = factory

    def export(self, port, name: str):
        """Expose an internal port as if it were a port on this Graph."""
        if name in self._exports:
            raise ValueError(f"export name '{name}' already used")
        self._exports[name] = port

    def _find_root_graph(self,edge):
        inport,outport = edge
        in_graph = inport.component.graph
        out_graph = outport.component.graph
        ingraph,inlevel = self._walk(in_graph)
        outgraph,outlevel = self._walk(out_graph)
        if inlevel>outlevel:
            return ingraph
        else:
            return outgraph


    def _walk(self, graph):
        level = 0
        cur = graph
        while cur.parent is not None:
            cur = cur.parent
            level += 1
        return cur, level


    def _add_subgraph(self, name, graph):
        new = graph.copy(name)
        new.is_subgraph = True
        new.__dict__['parent'] = self
        self.nodes = {**self.nodes, **new.nodes}
        self.edges += new.edges
        for node, data in new.networkx_graph.nodes(data=True):
            self.networkx_graph.add_node(node, **data)
        for u, v in new.networkx_graph.edges:
            self.networkx_graph.add_edge(u, v)
        self._local_namespace[name] = new
        self.__dict__[name]=new

    def copy(self,gname) -> "Graph":
        """
        Return a deep copy of this Graph, cloning all Components and Ports,
        rebuilding edges.
        """
        new = Graph()

        comp_map: dict[Component, Component] = {}
        new.nodes = {}
        for name, comp in self.nodes.items():
            name = f'{gname}_{name}'
            new_comp = comp.copy()
            new_comp.name = name
            new_comp.graph = new
            new.nodes[name] = new_comp
            comp_map[comp] = new_comp
            new.__dict__[name] = new_comp
            new._local_namespace[name] = new_comp

        # Clone component groups
        new.group_handles = {}
        for alias, group in self.group_handles.items():
            new_alias = f'{gname}_{alias}'
            new_group = ComponentGroup(group.comp_cls, group.count)
            new_group.name = new_alias
            new_group.graph = new
            for i in range(group.count):
                comp_name = f'{new_alias}_{i}'
                comp = new.nodes[comp_name]
                new_group.group.append(comp)
                comp.group_name = new_alias
            new.group_handles[new_alias] = new_group
            new.__dict__[new_alias] = new_group
            new._local_namespace[new_alias] = new_group

        new.edges = []
        new.networkx_graph = nx.DiGraph()
        for outp, inp in self.edges:
            cloned_out_comp = comp_map[outp.component]
            cloned_in_comp  = comp_map[inp.component]
            new_outp = getattr(cloned_out_comp, outp.name)
            new_inp  = getattr(cloned_in_comp,  inp.name)
            new_outp.connect(new_inp)

        new._exports = {}
        for export_name, port in self._exports.items():
            comp = port.component
            if isinstance(comp, ComponentGroup):
                alias = f'{gname}_{comp.name}'
                cloned_group = new.group_handles[alias]
                cloned_port = getattr(cloned_group, port.name)
            else:
                cloned_comp = comp_map[comp]
                cloned_port = getattr(cloned_comp, port.name)
            new._exports[export_name] = cloned_port

        return new


    def to_json(self):
        def encode_init(val):
            if isinstance(val, OutputPort):
                return {"component": val.component.name, "port": val.name}
            return val

        nodes = {}
        for name, comp in self.nodes.items():
            # Skip individual members of ComponentGroup; they are captured via
            # a single entry in ``group_handles`` below.
            if comp.group_name is not None or isinstance(comp, GroupConnector):
                continue

            info = {
                "module": comp.__class__.__module__,
                "class": comp.__class__.__name__,
                "group": comp.group_name,
                "inports": {
                    ip.name: {
                        "capacity": getattr(ip.channel, "capacity", None),
                        "initial": encode_init(ip.initialization_value),
                        "batch_size": ip.batch_size,
                    }
                    for ip in comp.inports
                },
            }
            nodes[name] = info

        edgeset = set()
        for o, i in self.edges:
            if isinstance(o.component, GroupConnector) or isinstance(i.component, GroupConnector):
                continue
            src = o.component.group_name or o.component.name
            dst = i.component.group_name or i.component.name
            edgeset.add((src, o.name, dst, i.name))

        for conn in (c for c in self.nodes.values() if isinstance(c, GroupConnector)):
            ups = [o for o, inp in self.edges if inp.component is conn]
            downs = [inp for outp, inp in self.edges if outp.component is conn]
            if not ups or not downs:
                continue
            src = ups[0].component.group_name or ups[0].component.name
            dst = downs[0].component.group_name or downs[0].component.name
            edgeset.add((src, ups[0].name, dst, downs[0].name))

        edges = [list(e) for e in sorted(edgeset)]

        groups = {}
        for alias, grp in self.group_handles.items():
            first = grp.group[0]
            groups[alias] = {
                "module": grp.comp_cls.__module__,
                "class": grp.comp_cls.__name__,
                "count": grp.count,
                "inports": {
                    ip.name: {
                        "capacity": getattr(ip.channel, "capacity", None),
                        "initial": encode_init(ip.initialization_value),
                        "batch_size": ip.batch_size,
                    }
                    for ip in first.inports
                },
            }

        exports = {}
        for name, port in self._exports.items():
            if isinstance(port.component, ComponentGroup):
                exports[name] = {"group": port.component.name, "port": port.name}
            else:
                exports[name] = {
                    "component": port.component.name,
                    "port": port.name,
                }

        subgraphs = {
            alias: sg.to_json()
            for alias, sg in self.__dict__.items()
            if isinstance(sg, Graph) and sg.is_subgraph
        }

        return {
            "nodes": nodes,
            "edges": edges,
            "group_handles": groups,
            "exports": exports,
            "subgraphs": subgraphs,
        }

    @classmethod
    def from_json(cls, data):
        import importlib

        def load_cls(mod, name):
            module = importlib.import_module(mod)
            return getattr(module, name)

        g = cls()

        # subgraphs first
        for alias, subdata in data.get("subgraphs", {}).items():
            sub = cls.from_json(subdata)
            sub.is_subgraph = True
            sub.parent = g
            g.nodes.update(sub.nodes)
            g.edges.extend(sub.edges)
            for node, data2 in sub.networkx_graph.nodes(data=True):
                g.networkx_graph.add_node(node, **data2)
            for u, v in sub.networkx_graph.edges:
                g.networkx_graph.add_edge(u, v)
            g.__dict__[alias] = sub
            g._local_namespace[alias] = sub

        # groups
        for alias, info in data.get("group_handles", {}).items():
            comp_cls = load_cls(info["module"], info["class"])
            group = ComponentGroup(comp_cls, info["count"])
            g._add_component_group(alias, group)

            for comp in group.group:
                for ip_name, ip_info in info.get("inports", {}).items():
                    ip = getattr(comp, ip_name)
                    val = ip_info.get("initial")
                    if isinstance(val, dict) and "component" in val:
                        src = g.nodes[val["component"]]
                        val = getattr(src, val["port"])
                    ip.initialization_value = val
                    ip.batch_size = ip_info.get("batch_size")
                    cap = ip_info.get("capacity")
                    if cap is not None:
                        ip.channel.capacity = cap

        # components
        for name, info in data.get("nodes", {}).items():
            if name in g.nodes:
                comp = g.nodes[name]
            else:
                comp_cls = load_cls(info["module"], info["class"])
                comp = comp_cls(name)
                g.node(name, comp)

            for ip_name, ip_info in info.get("inports", {}).items():
                ip = getattr(comp, ip_name)
                val = ip_info.get("initial")
                if isinstance(val, dict) and "component" in val:
                    src = g.nodes[val["component"]]
                    val = getattr(src, val["port"])
                ip.initialization_value = val
                ip.batch_size = ip_info.get("batch_size")
                cap = ip_info.get("capacity")
                if cap is not None:
                    ip.channel.capacity = cap

        for out_c, out_p, in_c, in_p in data.get("edges", []):
            if out_c in g.group_handles:
                src = getattr(g.group_handles[out_c], out_p)
            else:
                src = getattr(g.nodes[out_c], out_p)
            if in_c in g.group_handles:
                dst = getattr(g.group_handles[in_c], in_p)
            else:
                dst = getattr(g.nodes[in_c], in_p)
            src >> dst

        for name, info in data.get("exports", {}).items():
            if "group" in info:
                port = getattr(g.group_handles[info["group"]], info["port"])
            else:
                port = getattr(g.nodes[info["component"]], info["port"])
            g._exports[name] = port
        
        serialized_names = set(data.get("nodes", {}).keys())
        for alias, info in data.get("group_handles", {}).items():
            serialized_names.update(f"{alias}_{i}" for i in range(info.get("count", 0)))
        for subdata in data.get("subgraphs", {}).values():
            serialized_names.update(subdata.get("nodes", {}).keys())
            for alias, info in subdata.get("group_handles", {}).items():
                serialized_names.update(
                    f"{alias}_{i}" for i in range(info.get("count", 0))
                )
        extra = [
            n
            for n in list(g.nodes.keys())
            if n not in serialized_names and not isinstance(g.nodes[n], GroupConnector)
        ]
        for name in extra:
            comp = g.nodes.pop(name)
            if g.networkx_graph.has_node(comp):
                g.networkx_graph.remove_node(comp)

        return g

    def _break_cycles(self):
        """
        Breaks cycles in the graph for BFS coloring.
        """
        try:
            cycle = nx.find_cycle(self.networkx_graph, orientation='original')
            # Removing an edge to break the cycle
            self.networkx_graph.remove_edge(*cycle[0])
            return True
        except nx.NetworkXNoCycle:
            return False

    def check_deadlocks(self) -> None:
        """
        Static analysis for deadlocks. Raises DeadlockError if any of:
          1) A cycle of components with no seeded inport
          2) Any inport that has neither a default nor an upstream
          3) Any port used as a trigger (clk/sync) that has no upstream
          4) Any exported inport that is never wired
        """
        issues: list[str] = []

        # Build mapping: InputPort -> [outgoing OutputPort, ...]
        port_to_upstreams: dict[InputPort, list] = {}
        for outp, inp in self.edges:
            port_to_upstreams.setdefault(inp, []).append(outp)

        # 1) Cycles lacking a default seed:
        for scc in nx.strongly_connected_components(self.networkx_graph):
            if len(scc) > 1 or any(self.networkx_graph.has_edge(n, n) for n in scc):
                # ensure at least one component in the cycle has an inport.initialization_value
                seeded = False
                for comp_name in scc:
                    comp = self.nodes[comp_name]
                    for ip in comp.inports:
                        if ip.initialization_value is not None:
                            seeded = True
                            break
                if seeded:
                    break
                if not seeded:
                    issues.append(
                        f"Cycle deadlock: components {sorted(scc)} form a cycle with no default seed."
                    )

        # 2) Orphan inports (no default, no upstream):
        for comp_name, comp in self.nodes.items():
            for ip in comp.inports:
                if ip.initialization_value is None and not port_to_upstreams.get(ip):
                    issues.append(
                        f"Orphan port: '{comp_name}.{ip.name}' has no producer and no default."
                    )

        # 3) Unused clk/sync triggers:
        # comp._groups is a list of tuples of ports used as triggers
        for comp_name, comp in self.nodes.items():
            for group in comp._groups:
                for port in group:
                    if port in comp.inports and not port_to_upstreams.get(port):
                        issues.append(
                            f"Trigger deadlock: '{comp_name}.{port.name}' marked as clk/sync but never fed."
                        )

        # 4) Unconnected exported inports:
        for export_name, port in self._exports.items():
            if isinstance(port, InputPort) and not port_to_upstreams.get(port):
                issues.append(
                    f"Export deadlock: exported inport '{export_name}' has no upstream."
                )

        if issues:
            raise DeadlockError(issues)

    def shard(self, num_workers):
        """
        Shards the graph into GraphWorkers, using the advanced
        shard_graph_final2 strategy to assign each node to a worker.
        """
        # 1) Break all cycles so we can topologically sort
        while self._break_cycles():
            pass
    
        G = self.networkx_graph
    
        # 2) Build a helper map: group name → list of member nodes
        group_map = {}
        for node, data in G.nodes(data=True):
            grp = data.get('group')
            if grp:
                name = grp[0]
                group_map.setdefault(name, []).append(node)
    
        # 3) Topological sort
        topo = list(nx.topological_sort(G))
    
        # 4) Advanced worker‐assignment
        color_map = {}
        rr = 0
        for node in topo:
            # Unpack group info if any
            grp = G.nodes[node].get('group')
            if grp:
                name, idx, size = grp
                members = group_map[name]
    
                # (1) Parallel perfect‐matching: reuse parent’s worker
                preds = list(G.predecessors(node))
                if len(preds) == 1 and len(members) <= num_workers:
                    parents = [next(G.predecessors(m)) for m in members]
                    if len(set(parents)) == len(members):
                        color_map[node] = color_map[parents[idx]]
                        continue
    
            # (2) Split stage 1→N
            preds = list(G.predecessors(node))
            if len(preds) == 1:
                p = preds[0]
                children = list(G.successors(p))
                if len(children) > 1:
                    grp_c = G.nodes[node].get('group')
                    if grp_c and grp_c[2] == len(children):
                        i = children.index(node)
                        pw = color_map[p]
                        if len(children) < num_workers:
                            avail = [w for w in range(num_workers) if w != pw]
                            color_map[node] = avail[i % len(avail)]
                        else:
                            color_map[node] = i % num_workers
                        continue
    
            # (3) Broadcast stage M→workers
            assigned = False
            for p in preds:
                children = list(G.successors(p))
                if len(children) == num_workers:
                    i = children.index(node)
                    color_map[node] = i
                    assigned = True
                    break
            if assigned:
                continue
    
            # (4) Fallback: global round-robin
            color_map[node] = rr
            rr = (rr + 1) % num_workers
    
        # 5) Instantiate workers and assign components
        workers = [GraphWorker() for _ in range(num_workers)]
        for comp_name, comp in self.nodes.items():
            worker_index = color_map.get(comp_name, 0)
            workers[worker_index].components.append(comp)
        return workers

    def start(self, num_workers=None):
        """Starts the graph, without joining it.

        Args:
            num_workers (int, optional): If specified, spawns this amount of processes. If unspecified, uses either the cpu count or the total number of nodes, whichever is smaller.
        """
        if num_workers is None:
            num_workers = min(len(self.nodes), multiprocess.cpu_count())

        self.check_deadlocks()
        
        workers = self.shard(num_workers)
        self.processes = []
        for worker in workers:
            p = multiprocess.Process(target=worker.run)
            p.start()
            self.processes.append(p)

    def join(self):
        """Joins all processes in the graph. Handles KeyboardInterrupts gracefully.
        """
        try:
            for p in self.processes:
                p.join()
        except KeyboardInterrupt:
            log.warning("\nGraph execution cancelled by user.")
        finally:
            for p in self.processes:
                p.terminate()

    def run(self, num_workers=None):
        """Starts the graph and joins it

        Args:
            num_workers (int, optional): If specified, spawns this amount of processes. If unspecified, uses either the cpu count or the total number of nodes, whichever is smaller.
        """
        self.start(num_workers)
        self.join()


    def to_dot(self):
        """
        Dark‐themed GraphViz with HTML labels and bright red borders:
        - Inports on top
        - Component name in middle
        - Outports on bottom, with equal column spans
        """
        if not Digraph:
            log.warning('to_dot requires graphviz:\n    pip install graphviz')
            return

        dot = Digraph(
            comment='Cascata Graph (red-border)',
            format='png',
            graph_attr={'rankdir':'TD','bgcolor':'black', 'dpi':'300', 'nodesep':'1.0' ,'ranksep':'1.0'},
            node_attr={'shape':'none','margin':'0'},
            edge_attr={'color':'#FFFFFF'},
        )

        for name, comp in self.nodes.items():
            inps  = sorted(comp.inports,  key=lambda p: p.name)
            outs  = sorted(comp.outports, key=lambda p: p.name)
            n_in  = len(inps)
            n_out = len(outs)
            cols  = max(n_in, n_out, 1)


            in_cells = ''
            if n_in:
                for ip in inps:
                    in_cells += (
                        f'<TD PORT="{ip.name}" BORDER="1" COLOR="111111" '
                        'BGCOLOR="#2e2e2e" CELLPADDING="4" CELLSPACING="0">'
                        f'<FONT COLOR="white">{ip.name}</FONT>'
                        '</TD>'
                    )
                for _ in range(cols - n_in):
                    in_cells += (
                        '<TD BORDER="1" COLOR="111111" BGCOLOR="#2e2e2e" '
                        'CELLPADDING="4" CELLSPACING="0"></TD>'
                    )
            else:
                in_cells = (
                    f'<TD COLSPAN="{cols}" BORDER="1" COLOR="111111" '
                    'BGCOLOR="#2e2e2e" CELLPADDING="4" CELLSPACING="0"></TD>'
                )
            top_row = f'<TR>{in_cells}</TR>'


            mid_row = (
                f'<TR>'
                f'<TD COLSPAN="{cols}" BORDER="1" COLOR="111111" '
                'BGCOLOR="#444444" CELLPADDING="6" CELLSPACING="0">'
                f'<FONT COLOR="white"><B>{name}@{comp.__class__.__name__}</B></FONT>'
                '</TD>'
                '</TR>'
            )


            if n_out:
                base = cols // n_out
                rem  = cols - base * n_out
                out_cells = ''
                for i, op in enumerate(outs):
                    span = base + (1 if i == n_out - 1 and rem else 0)
                    out_cells += (
                        f'<TD PORT="{op.name}" COLSPAN="{span}" BORDER="1" COLOR="black" '
                        'BGCOLOR="#2e2e2e" CELLPADDING="4" CELLSPACING="0">'
                        f'<FONT COLOR="white">{op.name}</FONT>'
                        '</TD>'
                    )
            else:
                out_cells = (
                    f'<TD COLSPAN="{cols}" BORDER="1" COLOR="black" '
                    'BGCOLOR="#2e2e2e" CELLPADDING="4" CELLSPACING="0"></TD>'
                )
            bot_row = f'<TR>{out_cells}</TR>'

            html = (
                '<<TABLE BORDER="1" COLOR="black" CELLBORDER="0" CELLSPACING="0">\n'
                f'  {top_row}\n'
                f'  {mid_row}\n'
                f'  {bot_row}\n'
                '</TABLE>>'
            )
            dot.node(name, label=html)


        for outp, inp in self.edges:
            tail = f"{outp.component.name}:{outp.name}:s"
            head = f"{inp.component.name}:{inp.name}:n"
            dot.edge(tail, head)

        for name, comp in self.nodes.items():
            for ip in comp.inports:
                if ip.initialization_value is not None:
                    init = f'{name}.{ip.name}.init'
                    dot.node(
                        init,
                        label=f'< <FONT COLOR="white">{ip.initialization_value}</FONT> >',
                        shape='ellipse', style='filled',
                        fillcolor='#666666', fontcolor='white',
                        color='black'
                    )
                    dot.edge(init, f'{name}:{ip.name}')

        return dot

class GraphWorker:

    """Main worker for Cascata Graphs.
    Runs all run() methods of all components in the worker asynchronously.

    Attributes:
        components (list): List of components whos run methods will be spawned on this process
    """

    def __init__(self):
        self.components = []

    def run(self):
        """Creates 
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            loop.run_until_complete(asyncio.gather(
                *[comp.run() for comp in self.components]))
        except Exception as e:
            log.error(f"Exception in GraphWorker: {e}")
            traceback.print_exc()
        finally:
            loop.close()



import logging
import sys
import contextvars
import hashlib
from .component import Component

# ANSI escape codes
RESET = "\033[0m"
LEVEL_COLORS = {
    'debug': "\033[34m",   # blue
    'info': "\033[36m",    # cyan
    'warn': "\033[33m",    # yellow
    'warning': "\033[33m", # alias for warning
    'error': "\033[31m",   # red
}

# Generate a bright ANSI 256-color code deterministically from component class name
def _get_component_color(class_name: str) -> str:
    """
    Generate a bright, bold ANSI 256-color code deterministically from the component class name using hash.
    """
    digest = hashlib.md5(class_name.encode('utf-8')).digest()
    # Use two bytes to spread across the 216-color cube (starting at 16)
    idx = (digest[0] << 8 | digest[1]) % 216  # range 0-215
    code = 16 + idx  # ANSI 256-color code: 16-231 are the color cube
    return f"\033[38;5;{code}m"

# Context var to track current component
_current_component = contextvars.ContextVar('current_component', default=None)

# Monkey-patch Component.run to set context for logging
_original_run = Component.run
async def _run_with_context(self, *args, **kwargs):
    token = _current_component.set(self)
    try:
        return await _original_run(self, *args, **kwargs)
    finally:
        _current_component.reset(token)
Component.run = _run_with_context

class ContextAwareHandler(logging.StreamHandler):
    def emit(self, record):
        try:
            msg = record.getMessage()
            level = record.levelname.lower()
            level_color = LEVEL_COLORS.get(level, "")
            reset = RESET
            comp = _current_component.get()
            if comp is not None:
                comp_name = comp.group_name if getattr(comp, 'group_name', None) is not None else comp.name
                comp_class = comp.__class__.__name__
                comp_color = _get_component_color(comp_class)
                prefix = f"{level_color}{record.levelname}{reset} - {comp_name}{comp_color}@{comp_class}{reset}: "
            else:
                prefix = f"{level_color}{record.levelname}{reset} - Cascata: "
            stream = self.stream or sys.stdout
            stream.write(f"{prefix}{msg}\n")
            self.flush()
        except Exception:
            self.handleError(record)

# Configure module-level logger
log = logging.getLogger('cascata')
log.setLevel(logging.DEBUG)
log.handlers.clear()
handler = ContextAwareHandler()
log.addHandler(handler)
log.propagate = False



import asyncio
from collections import OrderedDict
from contextlib import AsyncExitStack
from aioitertools import zip as azip, chain, zip_longest, cycle
from copy import deepcopy
from typing import get_type_hints
from .channel import Channel
from .port import InputPort, OutputPort, PortHandler, PersistentValue

# Registry for component metadata
component_registry = {}


def _registry_insert(module_path, component_name, metadata):
    """Insert component metadata into the nested registry."""
    parts = module_path.split(".")
    node = component_registry
    for part in parts:
        node = node.setdefault(part, {})
    node[component_name] = metadata


class ComponentMeta(type):
    def __mul__(cls, count):
        if not issubclass(cls, Component) or not isinstance(count, int):
            return NotImplemented
        return ComponentGroup(cls, count)


class Component(metaclass=ComponentMeta):
    def __init__(self, name):
        self.name = name
        self.group_name = None
        self.ports = {}
        self.graph = None
        self.inports = set()
        self.outports = set()
        self._vals = OrderedDict()  # Holds references to ports and persistent values
        self._persists = OrderedDict()
        self._runner = None
        self._groups = []

    def __repr__(self):
        inports_text = " | ".join(port.name for port in self.inports)
        outports_text = " | ".join(port.name for port in self.outports)
        component_text = self.name

        # Determining the longest string for centering
        max_length = max(len(inports_text), len(component_text), len(outports_text))

        # Centering each line based on the longest string
        inports_text_centered = inports_text.center(max_length)
        component_text_centered = component_text.center(max_length)
        outports_text_centered = outports_text.center(max_length)

        return f"\n{'_'*(max_length+4)}\n| {inports_text_centered} |\n| {component_text_centered} |\n| {outports_text_centered} |\n{'¯'*(max_length+4)}"

    def __getattr__(self, attr):
        return self.ports.get(attr)

    def copy(self):

        new = self.__class__(self.name)

        for ip in self.inports:
            new_ip = getattr(new, ip.name)
            new_ip.initialization_value = ip.initialization_value
            new_ip.batch_size = ip.batch_size

        new._groups = [
            tuple(getattr(new, port.name) for port in group) for group in self._groups
        ]

        for name, pv in self._persists.items():
            new_pv = new._persists.get(name)
            new_pv.set_args(*pv._args, **pv._kwargs)
            if pv._initialized:
                new_pv.set(pv.value)

        return new

    def add_inport(self, port):
        self.inports.add(port)
        port.component = self
        self.ports[port.name] = port
        self._vals[port] = None

    def add_outport(self, port):
        self.outports.add(port)
        port.component = self
        self.ports[port.name] = port
        self._vals[port] = port

    def add_persist(self, name, initializer):
        pv = PersistentValue(initializer)
        self._persists[name] = pv
        self.ports[name] = pv
        self._vals[pv] = pv
        return pv

    def set_runner(self, runner_function):
        self._runner = runner_function

    @classmethod
    def test(cls, **kwargs):
        """Execute the component's runner with provided inport values.

        Keyword arguments must supply a value for every inport of the
        component.  Outports are replaced with dummy ports that simply
        collect the items sent to them.  The collected items are returned
        as a ``dict`` keyed by outport name.

        Example
        -------
        >>> @component
        ... @inport('x')
        ... @outport('y')
        ... async def dup(x, y):
        ...     await y.send(x)
        >>> dup.test(x=1)
        {'y': [1]}
        """

        instance = cls(cls.__name__)

        class _Collector:
            def __init__(self):
                self.items = []

            async def send(self, item):
                self.items.append(item)

        outputs = {}
        args = []
        for port in instance._vals.keys():
            if isinstance(port, InputPort):
                if port.name not in kwargs:
                    raise TypeError(f"Missing value for inport '{port.name}'")
                args.append(kwargs[port.name])
            elif isinstance(port, OutputPort):
                collector = _Collector()
                outputs[port.name] = collector
                args.append(collector)
            elif isinstance(port, PersistentValue):
                port.get()
                args.append(port)

        asyncio.run(instance._runner(*args))

        return {name: collector.items for name, collector in outputs.items()}

    def clk(self, *ports):
        for port in ports:
            self._groups.append((port,))

    def sync(self, *ports):
        self._groups.append((ports))

    async def _group_listener(self, group, execute_runner):
        if execute_runner:

            async def listener(group):
                async for values in azip(*group):
                    self._vals.update(zip(group, values))
                    args = []
                    for v in self._vals.values():
                        if isinstance(v, PersistentValue):
                            v.get()
                        args.append(v)
                    await self._runner(*args)

        else:

            async def listener(group):
                async for values in azip(*group):
                    self._vals.update(zip(group, values))

        await listener(group)

    async def run(self):

        async with AsyncExitStack() as stack:
            await asyncio.gather(
                *[stack.enter_async_context(port.open()) for port in self.outports]
            )
            # Categorize ports based on initialization values
            initports = {
                port for port in self.inports if port.initialization_value is not None
            }

            # Initialize ports with initialization values
            for port in initports:
                if isinstance(port.initialization_value, OutputPort):
                    async for value in port:
                        self._vals[port] = value
                        break
                else:
                    self._vals[port] = port.initialization_value

            # Check if all ports are initports
            if initports == self.inports:
                # If all inports are initports, call the runner and exit
                await self._runner(*self._vals.values())
                return

            if not self._groups:
                self._groups.append(
                    [port for port in self.inports if port.initialization_value is None]
                )

            all_grouped_ports = set().union(*self._groups)
            exec_flags = [True for port in self._groups]
            ungrouped_ports = self.inports - all_grouped_ports - initports

            # Add each ungrouped port as a group on its own
            for port in ungrouped_ports:
                self._groups.append({port})
                exec_flags.append(False)
            # Prepare and run all group listeners concurrently
            group_listeners = [
                self._group_listener(group, flag)
                for group, flag in zip(self._groups, exec_flags)
            ]
            await asyncio.gather(*group_listeners)


def inport(port_name, capacity=10, default=None):
    def decorator(func):
        if not hasattr(func, "_inports"):
            func._inports = []
        func._inports.append((port_name, capacity, default))
        if not hasattr(func, "_val_decls"):
            func._val_decls = []
        func._val_decls.insert(0, ("inport", port_name, capacity, default))
        return func

    return decorator


def outport(port_name):
    def decorator(func):
        if not hasattr(func, "_outports"):
            func._outports = []
        func._outports.append(port_name)
        if not hasattr(func, "_val_decls"):
            func._val_decls = []
        func._val_decls.insert(0, ("outport", port_name))
        return func

    return decorator


def persist(name, initializer):
    def decorator(func):
        if not hasattr(func, "_persists"):
            func._persists = []
        func._persists.append((name, initializer))
        if not hasattr(func, "_val_decls"):
            func._val_decls = []
        func._val_decls.insert(0, ("persist", name, initializer))
        return func

    return decorator


def component(func):
    annotations = get_type_hints(func)

    class ComponentSubclass(Component):
        def __init__(self, *args, **kwargs):
            super().__init__(func.__name__)
            self._persists = OrderedDict()
            for decl in getattr(func, "_val_decls", []):
                kind = decl[0]
                if kind == "inport":
                    _, name, capacity, default = decl
                    port = InputPort(name, capacity, default)
                    port.type = annotations.get(name)
                    self.add_inport(port)
                elif kind == "outport":
                    _, name = decl
                    port = OutputPort(name)
                    port.type = annotations.get(name)
                    self.add_outport(port)
                elif kind == "persist":
                    _, name, init = decl
                    self.add_persist(name, init)
            self.set_runner(deepcopy(func))

    ComponentSubclass.__name__ = func.__name__
    ComponentSubclass.__module__ = func.__module__
    # Register component metadata
    # Use the explicit _inports and _outports declarations captured by the
    # decorators so that the registry retains the port names as well as their
    # properties (e.g. capacity, default value and type annotations).  Prior to
    # this the registry only stored bare type information which meant port names
    # were effectively dropped from the registry making introspection
    # impossible.
    inports_meta = {}
    for name, capacity, default in getattr(func, "_inports", []):
        inports_meta[name] = {
            "type": annotations.get(name),
            "capacity": capacity,
            "default": default,
        }

    outports_meta = {}
    for name in getattr(func, "_outports", []):
        outports_meta[name] = {"type": annotations.get(name)}

    metadata = {
        "module": func.__module__,
        "inports": inports_meta,
        "outports": outports_meta,
    }
    _registry_insert(func.__module__, func.__name__, metadata)

    return ComponentSubclass


class ComponentGroup:
    def __init__(self, comp_cls, count):
        self.comp_cls = comp_cls  # the Component subclass
        self.count = count  # number of copies
        # these get filled in by Graph._add_component_group
        self.name = None
        self.graph = None
        self.group = []  # list of Component instances
        self.ports = {}

    def __repr__(self):
        return f"<ComponentGroup {self.name} x{self.count}>"

    def __getattr__(self, attr):
        if not self.ports:
            for k in self.group[0].ports.keys():
                self.ports[k] = PortHandler(k, self)
        return self.ports[attr]

    def initialize(self, name, val):
        for comp in self.group:
            getattr(comp, name).__lt__(val)

    def connect(self, src, target):
        """
        Connect this group’s output-port `name` into `target`.
        - If `target` is a ComponentGroup of the same size: do 1:1 parallel wiring.
        - Otherwise, insert a M→N GroupConnector under a synthetic name.
        """

        if isinstance(src.component, ComponentGroup):
            M = src.component.count
            src_group = src.component.group
        else:
            src_group = [src.component]
            M = 1

        if isinstance(target.component, ComponentGroup):
            N = target.component.count
            dst_group = target.component.group
        else:
            dst_group = [target.component]
            N = 1

        # Case A: same-size group→group → straight 1:1
        if M == N:
            for srcg, dstg in zip(enumerate(self.group), enumerate(dst_group)):
                i, src_comp = srcg
                j, dst_comp = dstg
                src_comp.ports[src.name] >> dst_comp.ports[target.name]
            return

        # Case B/C: M→N with M != N → insert a bridge
        conn_name = (
            f"{src.component.name}.{src.name}-to-{target.component.name}.{target.name}"
        )
        # create connector component with base port name `name`
        bridge = GroupConnector(conn_name, target.name, M, N)
        # register it in the graph
        self.graph.node(conn_name, bridge)
        # 1) wire each of the M group members into bridge.in_{i}
        for i, src_comp in enumerate(src_group):
            src_comp.ports[src.name] >> bridge.ports[f"{target.name}_{i}_in"]

        # 2) wire each of the N bridge.out_{j} into the dst components
        for j, dst_comp in enumerate(dst_group):
            bridge.ports[f"{target.name}_{j}_out"] >> dst_comp.ports[target.name]

    def connect_batch(self, src, target, size=0):
        """Connect like :py:meth:`connect` but with batching semantics."""
        if isinstance(target, PortHandler):
            if isinstance(target.component, ComponentGroup):
                for comp in target.component.group:
                    getattr(comp, target.name).batch_size = size
            else:
                getattr(target.component, target.name).batch_size = size
        elif isinstance(target, InputPort):
            target.batch_size = size
        else:
            raise TypeError("target must be InputPort or PortHandler")
        self.connect(src, target)


class GroupConnector(Component):
    def __init__(self, base, name, M, N):
        self.name = base
        self.port_name = name
        super().__init__(base)
        self.M = M
        self.N = N
        for i in range(M):
            self.add_inport(InputPort(f"{name}_{i}_in"))
        for j in range(N):
            self.add_outport(OutputPort(f"{name}_{j}_out"))

    def copy(self):
        return GroupConnector(self.name, self.port_name, self.M, self.N)

    async def run(self):
        async with AsyncExitStack() as stack:
            await asyncio.gather(
                *[stack.enter_async_context(port.open()) for port in self.outports]
            )
            in_iters = [
                self.ports[f"{self.port_name}_{i}_in"].__aiter__()
                for i in range(self.M)
            ]
            merged = chain.from_iterable(
                zip_longest(*in_iters, fillvalue=StopAsyncIteration)
            )
            dist = cycle(self.outports)
            async for item in merged:
                if item is StopAsyncIteration:
                    break
                thisport = await anext(dist)
                await thisport.send(item)



import asyncio
import importlib
import sys
import multiprocess as multiprocessing
import networkx as nx

import pytest
from cascata import component, inport, outport, Graph
from cascata.port import PersistentValue
from cascata.log import log

@component
@outport('o')
async def Prod(o):
    await o.send(1)

@component
@inport('i')
async def Cons(i):
    pass

@component
@inport('a')
@inport('b')
async def TwoInputs(a, b):
    pass

@component
@inport('i', default=1)
@outport('o')
async def Seeded(i, o):
    await o.send(i)

@component
@inport('i')
@outport('o')
async def Pipe(i, o):
    await o.send(i)

@component
@inport('i')
@outport('o1')
@outport('o2')
async def ManyOut(i, o1, o2):
    await o1.send(i)
    await o2.send(i)

def test_persistent_value_extra_paths():
    pv = PersistentValue(lambda a, b=0: [a, b])
    pv < (1, 2)
    assert pv.get() == [1, 2]
    pv < {'a': 5, 'b': 3}
    assert pv.get() == [5, 3]

def test_sync_and_listener_no_runner():
    c = TwoInputs('c')
    c.sync(c.b)
    async def send_and_listen():
        async with c.a.channel.open():
            await c.a.channel.put(42)
        await c._group_listener((c.a,), False)
        assert c._vals[c.a] == 42
    asyncio.run(send_and_listen())


def test_logging_emit_error(monkeypatch):
    handler = log.handlers[0]
    called = []
    monkeypatch.setattr(handler, 'handleError', lambda record: called.append(True))
    class BadStream:
        def write(self, msg):
            raise ValueError
    monkeypatch.setattr(handler, 'stream', BadStream())
    log.info('boom')
    assert called


def test_run_initialization_from_output():
    @component
    @outport('o')
    async def Src(o):
        await o.send(7)

    @component
    @inport('i')
    async def Dest(i):
        pass

    g = Graph()
    g.s = Src
    g.d = Dest
    g.s.o >> g.d.i
    g.d.i.initialization_value = g.s.o

    async def run_pair():
        async with asyncio.TaskGroup() as tg:
            tg.create_task(g.s.run())
            tg.create_task(g.d.run())

    asyncio.run(run_pair())


def test_run_adds_ungrouped_ports():
    @component
    @inport('a')
    @inport('b')
    async def Comp(a, b):
        pass

    c = Comp('x')
    c.clk(c.b)

    async def feed():
        async with c.a.channel.open():
            await c.a.channel.put(1)

    async def run_all():
        async with asyncio.TaskGroup() as tg:
            tg.create_task(feed())
            async with c.b.channel.open():
                await c.b.channel.put(2)
            tg.create_task(c.run())

    asyncio.run(run_all())


def test_to_dot_padding():
    g = Graph()
    g.m = ManyOut
    from cascata.graph import Digraph
    dot = g.to_dot()
    if Digraph:
        assert 'digraph' in dot.source
    else:
        assert dot is None


def test_graph_setattr_instance():
    g = Graph()
    inst = Prod('inst')
    g.inst = inst
    assert g.inst is inst and 'inst' in g.nodes


def test_find_root_graph_branch():
    sub = Graph()
    sub.c = Cons
    sub.export(sub.c.i, 'inp')
    g = Graph()
    g.sub = sub
    g.p = Prod
    g.p.o >> g.sub.inp
    assert (g.p.o, g.sub.inp) in g.edges


def test_break_cycles_true(monkeypatch):
    g = Graph()
    g.a = Pipe
    g.b = Pipe
    g.a.o >> g.b.i
    g.b.o >> g.a.i
    monkeypatch.setattr(nx, 'find_cycle', lambda G, orientation='original': [(g.a, g.b)])
    assert g._break_cycles() is True
    assert (g.a, g.b) not in g.networkx_graph.edges


def test_check_deadlocks_seeded_cycle():
    g = Graph()
    g.a = Pipe
    g.b = Pipe
    g.a.o >> g.b.i
    g.b.o >> g.a.i
    g.a.i.initialize(1)
    g.nodes[g.a] = g.a
    g.nodes[g.b] = g.b
    g.check_deadlocks()


def test_shard_branches():
    g1 = Graph()
    g1.p1 = Prod
    g1.p2 = Prod
    g1.cons = Cons * 2
    g1.p1.o >> g1.cons.group[0].i
    g1.p2.o >> g1.cons.group[1].i
    calls = iter([True, False])
    def fake_break():
        return next(calls)
    g1._break_cycles = fake_break
    g1.shard(2)

    g2 = Graph()
    g2.p = Prod
    g2.cons = Cons * 2
    g2.p.o >> g2.cons.group[0].i
    g2.p.o >> g2.cons.group[1].i
    g2.shard(3)


def test_join_keyboard_interrupt(monkeypatch, caplog):
    class KIProcess(multiprocessing.Process):
        def join(self):
            super().join()
            raise KeyboardInterrupt
    caplog.set_level('WARNING')
    g = Graph()
    g.p = Prod
    handler = log.handlers[0]
    from io import StringIO
    stream = StringIO()
    monkeypatch.setattr(handler, 'stream', stream)
    monkeypatch.setattr(multiprocessing, 'Process', KIProcess)
    g.run(1)
    assert 'Graph execution cancelled' in stream.getvalue()


def test_reload_without_graphviz(monkeypatch):
    import cascata.graph as gmod
    saved = sys.modules.pop('graphviz', None)
    sys.modules['graphviz'] = None
    importlib.reload(gmod)
    assert gmod.Digraph is False
    gmod.Graph().to_dot()
    if saved is not None:
        sys.modules['graphviz'] = saved
    else:
        sys.modules.pop('graphviz', None)
    importlib.reload(gmod)


def test_start_default_workers(monkeypatch):
    g = Graph()
    g.p = Prod
    monkeypatch.setattr(multiprocessing, 'cpu_count', lambda: 1)
    g.start()
    g.join()



import asyncio
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import pytest
from cascata import Graph, inport, outport, component
from cascata.graph import DeadlockError, Digraph
from cascata.component import GroupConnector

@component
@inport('inp')
@outport('out')
async def Echo(inp, out):
    await out.send(inp)

@component
@inport('inp')
async def Sink(inp):
    pass


def test_component_repr():
    comp = Echo('e')
    text = repr(comp)
    # representation should include port names and component class name
    assert 'inp' in text and 'out' in text and 'Echo' in text


def test_deadlock_cycle():
    g = Graph()
    g.a = Echo
    g.b = Echo
    g.a.out >> g.b.inp
    g.b.out >> g.a.inp
    import networkx as nx
    g.networkx_graph = nx.DiGraph()
    for outp, inp in g.edges:
        g.networkx_graph.add_edge(outp.component.name, inp.component.name)
    with pytest.raises(DeadlockError) as exc:
        g.check_deadlocks()
    assert 'Cycle deadlock' in str(exc.value)


def test_deadlock_orphan_port():
    g = Graph()
    g.sink = Sink
    with pytest.raises(DeadlockError) as exc:
        g.check_deadlocks()
    assert "Orphan port" in str(exc.value)


def test_deadlock_trigger():
    g = Graph()
    g.e = Echo
    g.e.clk(g.e.inp)
    with pytest.raises(DeadlockError) as exc:
        g.check_deadlocks()
    assert 'Trigger deadlock' in str(exc.value)


def test_deadlock_unconnected_export():
    g = Graph()
    g.e = Echo
    g.export(g.e.inp, 'ext')
    with pytest.raises(DeadlockError) as exc:
        g.check_deadlocks()
    assert 'Export deadlock' in str(exc.value)


def test_to_dot_without_graphviz():
    g = Graph()
    result = g.to_dot()
    if Digraph:
        assert 'digraph' in result.source
    else:
        assert result is None


def test_shard_and_group_connector():
    g = Graph()
    g.prods = Echo * 2
    g.sink = Sink
    g.prods.out >> g.sink.inp
    workers = g.shard(2)
    # Connector node should exist
    assert any(isinstance(c, GroupConnector) for c in g.nodes.values())
    # All components assigned across two workers
    total = sum(len(w.components) for w in workers)
    assert total == len(g.nodes)



import asyncio
import multiprocess as multiprocessing
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from cascata import Graph, inport, outport, component, persist

# Define simple components for tests
@component
@inport('count', default=5)
@outport('out')
async def Producer(count, out):
    for i in range(count):
        await out.send(i)

@component
@inport('inp')
@outport('out')
async def Forward(inp, out):
    await out.send(inp)

@component
@inport('item')
@persist('store', lambda lst: lst)
async def Collector(item, store):
    store.get().append(item)


def test_component_creation():
    p = Producer('prod')
    c = Collector('c')
    assert {ip.name for ip in p.inports} == {'count'}
    assert {op.name for op in p.outports} == {'out'}
    assert 'store' in c.ports


def test_graph_creation_and_run():
    mgr = multiprocessing.Manager()
    result = mgr.list()

    g = Graph()
    g.producer = Producer
    g.forward = Forward
    g.collector = Collector

    g.collector.store < result
    g.producer.out >> g.forward.inp
    g.forward.out >> g.collector.item

    g.run(1)

    assert list(result) == list(range(5))


def test_graph_with_subgraph():
    mgr = multiprocessing.Manager()
    result = mgr.list()

    sub = Graph()
    sub.p = Producer
    sub.export(sub.p.out, 'out')

    g = Graph()
    g.sub = sub
    g.collector = Collector
    g.collector.store < result

    g.sub.out >> g.collector.item

    g.run(1)

    assert list(result) == list(range(5))


def test_grouped_components():
    mgr = multiprocessing.Manager()
    result = mgr.list()

    g = Graph()
    g.prods = Producer * 2
    g.colls = Collector * 2
    g.colls.store < result

    g.prods.out >> g.colls.item
    g.prods.initialize('count', 2)

    g.run(1)

    assert sorted(result) == [0, 0, 1, 1]


def test_subgraph_and_grouped_components():
    mgr = multiprocessing.Manager()
    result = mgr.list()

    sub = Graph()
    sub.p = Producer
    sub.export(sub.p.out, 'out')

    g = Graph()
    g.sub = sub
    g.colls = Collector * 2
    g.colls.store < result

    g.sub.out >> g.colls.item
    g.colls.store < result

    g.run(1)

    assert sorted(result) == list(range(5))


def test_subgraph_with_group_inside():
    mgr = multiprocessing.Manager()
    result = mgr.list()

    sub = Graph()
    sub.prods = Producer * 2
    sub.merger = Forward
    sub.prods.out >> sub.merger.inp
    sub.export(sub.merger.out, 'out')

    sub.prods.initialize('count', 2)

    g = Graph()
    g.sub = sub
    g.collector = Collector
    g.collector.store < result

    g.sub.out >> g.collector.item
    g.producer = Producer
    g.producer.count < 0  # avoid extra output

    g.run(1)

    assert sorted(result) == [0, 0, 1, 1]


def test_subgraph_exporting_group_port():
    mgr = multiprocessing.Manager()
    result = mgr.list()

    sub = Graph()
    sub.prods = Producer * 2
    sub.export(sub.prods.out, 'out')
    sub.prods.initialize('count', 2)

    g = Graph()
    g.sub = sub
    g.colls = Collector * 2
    g.colls.store < result

    g.sub.out >> g.colls.item

    g.run(1)

    assert sorted(result) == [0, 0, 1, 1]



import asyncio
import os
import sys
import logging
import tempfile
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import pytest
from cascata import Graph, inport, outport, component, persist
from cascata.graph import GraphWorker, DeadlockError, Digraph
from cascata.port import InputPort, OutputPort, PortHandler, PersistentValue
from cascata.log import log, _get_component_color

# Simple components
@component
@outport('o')
async def Prod(o):
    await o.send(1)

@component
@inport('i')
async def Cons(i):
    pass

@component
@inport('i')
@outport('o')
async def Pipe(i, o):
    await o.send(i)

@component
@inport('i1')
@inport('i2')
@outport('o')
async def Combiner(i1, i2, o):
    await o.send((i1, i2))

@component
async def Fail():
    raise RuntimeError('boom')

@component
async def Logger():
    log.info('hi')


def test_get_component_color():
    # color code should start with escape sequence
    c = _get_component_color('MyComp')
    assert c.startswith('\x1b[38;5;')

def test_persistent_value_lt_tuple():
    pv = PersistentValue(lambda x,y: x+y)
    pv < ((1,2), {})
    assert pv.get() == 3

def test_component_mul_invalid():
    with pytest.raises(TypeError):
        Prod * 2.5


def test_component_copy_persist():
    @component
    @persist('p', list)
    async def P(p):
        pass
    inst = P('x')
    inst.p < [1,2]
    inst.p.get()
    copied = inst.copy()
    assert copied.p.get() == [1,2]


def test_group_repr_and_port_handler():
    g = Graph()
    g.group = Prod * 2
    group = g.group
    text = repr(group)
    assert str(group.count) in text
    # access port via PortHandler
    ph = group.o
    assert isinstance(ph, PortHandler)


def test_port_lt_and_repr():
    g = Graph()
    g.p = Pipe
    g.c = Cons
    ip = g.c.i
    op = g.p.o
    repr(ip)
    repr(op)
    ip < 1
    op >> ip


def test_graph_attribute_error_and_export():
    g = Graph()
    g.p = Prod
    with pytest.raises(AttributeError):
        g.missing
    with pytest.raises(ValueError):
        g.export(g.p.o, 'x')
        g.export(g.p.o, 'x')


def test_break_cycles_and_shard():
    g = Graph()
    g.a = Pipe
    g.b = Pipe
    g.a.o >> g.b.i
    g.b.o >> g.a.i
    with pytest.raises(TypeError):
        g._break_cycles()
    g2 = Graph()
    g2.p1 = Pipe
    g2.p2 = Pipe
    g2.p1.o >> g2.p2.i
    workers = g2.shard(2)
    assert len(workers) == 2


def test_graph_to_dot_with_graphviz():
    g = Graph()
    g.p = Prod
    g.c = Cons
    g.p.o >> g.c.i
    dot = g.to_dot()
    if Digraph:
        assert 'digraph' in dot.source
    else:
        assert dot is None

def test_to_dot_initial_values():
    @component
    @outport('o')
    async def Src(o):
        await o.send(5)

    @component
    @inport('inp')
    async def Tgt(inp):
        pass

    g = Graph()
    g.src = Src
    g.tgt = Tgt
    g.tgt.inp < 1
    dot = g.to_dot()
    if Digraph:
        assert 'init' in dot.source
    else:
        assert dot is None


def test_logging_context_and_error(monkeypatch):
    g = Graph()
    g.l = Logger
    handler = log.handlers[0]
    with tempfile.TemporaryFile(mode='w+') as stream:
        monkeypatch.setattr(handler, 'stream', stream)
        g.run(1)
        stream.seek(0)
        assert 'Logger' in stream.read()



def test_graphworker_exception_logging():
    handler = log.handlers[0]
    old_stream = handler.stream
    with tempfile.TemporaryFile(mode='w+') as stream:
        handler.stream = stream
        g = Graph()
        g.f = Fail
        g.run(1)
        stream.seek(0)
        content = stream.read()
    handler.stream = old_stream
    assert 'Exception in GraphWorker' in content

